try:
    import matplotlib.pyplot as plt
except ImportError:
    import matplotlib

    matplotlib.use("agg")
    import matplotlib.pyplot as plt


def plot_histogram(x, x0, outfile, side):
    ax = plt.hist(x, bins=25)
    plt.axvline(x0, c="r", alpha=0.5, linestyle="--")
    plt.xlabel(
        "{}ifference in test statistic".format("Absolute d" if side == "both" else "D")
    )
    plt.ylabel("Frequency")
    plt.savefig(outfile)
    plt.close()



__author__ = "David J. Skelton"
__copyright__ = "Copyright 2019, David J. Skelton"
__license__ = "MIT"
__version__ = "0.1.8"
__maintainer__ = "David J. Skelton"
__email__ = "d.j.skelton1@gmail.com"
__status__ = "Production"

from collections import namedtuple as _nt

_GT = {"g", "gt", ">", "greater"}
_LT = {"l", "lt", "<", "lower"}
_BOTH = {"both"}
_RESULT = _nt("Result", ["lower", "upper", "confidence"])

from mcpt.permutation_test import permutation_test
from mcpt.correlation_permutation_test import correlation_permutation_test



import random as _rd
import multiprocessing as _mp

import numpy as _np

from mcpt import _GT, _LT, _BOTH, _RESULT
from mcpt.ci import wilson
from mcpt.plot import plot_histogram


# Handlers for the different sides.
def _greater(tup):
    combined, x_len, reference, f, seed = tup
    _rd.seed(seed)
    _rd.shuffle(combined)
    x, y = combined[:x_len], combined[x_len:]
    diff = f(x) - f(y)
    return diff >= reference, diff


def _lower(tup):
    combined, x_len, reference, f, seed = tup
    _rd.seed(seed)
    _rd.shuffle(combined)
    x, y = combined[:x_len], combined[x_len:]
    diff = f(x) - f(y)
    return diff <= reference, diff


def _both(tup):
    combined, x_len, reference, f, seed = tup
    _rd.seed(seed)
    _rd.shuffle(combined)
    x, y = combined[:x_len], combined[x_len:]
    diff = abs(f(x) - f(y))
    return diff >= reference, diff


def _job_hander(f, jobs, cores):
    if cores == 1:
        return map(f, jobs)

    with _mp.Pool(cores) as p:
        results = p.map(f, jobs)
    return results


# Main function.
def permutation_test(
    x, y, f, side, n=10000, confidence=0.99, plot=None, cores=1, seed=None
):
    if seed:
        rng = _rd.Random(seed)
    else:
        rng = _rd.Random()

    x_len = len(x)

    combined = list(x) + list(y)

    if callable(f):
        _f = f
    elif f == "mean":
        _f = _np.mean
    elif f == "median":
        _f = _np.median
    elif f == "stdev":
        _f = _np.std
    else:
        raise ValueError(
            "{} not valid for f -- must be a function, 'mean', 'median' or 'stdev'".format(
                f
            )
        )

    if side in _GT:
        diff = _f(x) - _f(y)
    elif side in _LT:
        diff = _f(x) - _f(y)
    elif side in _BOTH:
        diff = abs(_f(x) - _f(y))
    else:
        raise ValueError(
            "{} not valid for side -- should be 'greater', 'lower', or 'both'".format(
                side
            )
        )

    jobs = ((combined, x_len, diff, _f, rng.randint(0, 1e100)) for _ in range(n))

    if side in _GT:
        result = _job_hander(_greater, jobs, cores)
    elif side in _LT:
        result = _job_hander(_lower, jobs, cores)
    else:
        result = _job_hander(_both, jobs, cores)

    v = []
    p = 0

    for truth, val in result:
        p += truth
        v.append(val)

    p /= n

    if plot:
        plot_histogram(x=v, x0=diff, outfile=plot, side=side)

    lower, upper = wilson(p, n, confidence)
    return _RESULT(lower, upper, confidence)



import random as _rd
import multiprocessing as _mp

import scipy.stats as _st

from mcpt import _GT, _LT, _BOTH, _RESULT
from mcpt.ci import wilson
from mcpt.plot import plot_histogram


def _correlation_greater(tup):
    x, y, stat_0, f, seed = tup
    _rd.seed(seed)
    _rd.shuffle(y)
    stat = f(x, y)
    return stat >= stat_0, stat


def _correlation_lower(tup):
    x, y, stat_0, f, seed = tup
    _rd.seed(seed)
    _rd.shuffle(y)
    stat = f(x, y)
    return stat <= stat_0, stat


def _correlation_both(tup):
    x, y, stat_0, f, seed = tup
    _rd.seed(seed)
    _rd.shuffle(y)
    stat = abs(f(x, y))
    return stat >= stat_0, stat


def _pearsonr(x, y):
    return _st.pearsonr(x, y)[0]


def _spearmanr(x, y):
    return _st.spearmanr(x, y)[0]


def _job_hander(f, jobs, cores):
    if cores == 1:
        return map(f, jobs)

    with _mp.Pool(cores) as p:
        results = p.map(f, jobs)

    return results


def correlation_permutation_test(
    x, y, f, side, n=10000, confidence=0.99, plot=None, cores=1, seed=None
):
    """This function carries out Monte Carlo permutation tests comparing whether the correlation between two variables is statistically significant

    :param x: An iterable of X values observed
    :param y: An iterable of Y values observed
    :param f: The function for calculating the relationship strength between X and Y
    :param side: The side to use for hypothesis testing
    :param n: The number of permutations to sample, defaults to 10000
    :type n: int, optional
    :param confidence: The probability that the true p-value is contained in the intervals returned, defaults to 0.99
    :type confidence: float, optional
    :param plot: The name of a file to draw a plot of permuted correlations to, defaults to None
    :type plot: str, optional
    :param cores: The number of logical CPUs to use, defaults to 1
    :type cores: int, optional
    :param seed: The seed for randomisation, defaults to None
    :type seed: int, optional
    :return: Named tuple containing upper and lower bounds of p-value at the given confidence
    """

    if seed:
        rng = _rd.Random(seed)
    else:
        rng = _rd.Random()

    if callable(f):
        _f = f
    elif f == "pearsonr":
        _f = _pearsonr
    elif f == "spearmanr":
        _f = _spearmanr
    else:
        raise ValueError(
            "{} not valid for f -- must be a function, 'pearsonr', or 'spearmanr'".format(
                f
            )
        )

    _x = list(x)
    _y = list(y)

    if side in _GT:
        stat_0 = _f(_x, _y)
    elif side in _LT:
        stat_0 = _f(_x, _y)
    elif side in _BOTH:
        stat_0 = abs(_f(_x, _y))
    else:
        raise ValueError(
            "{} not valid for side -- should be 'greater', 'lower', or 'both'".format(
                side
            )
        )

    jobs = ((_x[:], _y[:], stat_0, _f, rng.randint(0, 1e100)) for _ in range(n))

    if side in _GT:
        result = _job_hander(_correlation_greater, jobs, cores)
    elif side in _LT:
        result = _job_hander(_correlation_lower, jobs, cores)
    else:
        result = _job_hander(_correlation_both, jobs, cores)

    v = []
    p = 0

    for truth, val in result:
        p += truth
        v.append(val)

    p /= n

    if plot:
        plot_histogram(x=v, x0=stat_0, outfile=plot, side=side)

    lower, upper = wilson(p, n, confidence)
    return _RESULT(lower, upper, confidence)



import scipy.stats as _st

# Functions for calculating probability distribution for
# the result.


def wilson(p, n, alpha):
    # Convert alpha to Z-score.
    # We want 1 - alpha proportion of values contained.
    z = _st.norm.ppf((1 - alpha) / 2 + alpha)
    # Calculate lower and upper bounds
    lower = (p + z * z / (2 * n) - z * ((p * (1 - p) + z * z / (4 * n)) / n) ** 0.5) / (
        1 + z * z / n
    )
    upper = (p + z * z / (2 * n) + z * ((p * (1 - p) + z * z / (4 * n)) / n) ** 0.5) / (
        1 + z * z / n
    )

    return (lower, upper)



import setuptools

# Read the contents of the README file
import io
from os import path

this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="mcpt",
    version="0.1.8",
    description="A Python library for calculating p-values using Monte Carlo sampling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ravenlocke/mcpt",
    author="David J. Skelton",
    author_email="d.j.skelton1@gmail.com",
    license="MIT",
    packages=setuptools.find_packages(),
    install_requires=["scipy", "numpy", "matplotlib"],
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires="~=3.5",
)
